# pylint: disable=wildcard-import
"""Video action recognition."""
from __future__ import absolute_import
from .vgg16_ucf101 import *
from .inceptionv3_ucf101 import *
from .inceptionv3_kinetics400 import *
